/***********************************************************************/
/*                                                                     */
/*                 HNFLLL (Havas, Majewski, Mathews)                   */
/*                                                                     */
/***********************************************************************/
#include <genpari.h>

static GEN **lambda, *D;

static GEN  mynegi(GEN x);
static long findi(GEN M);
static void Minus(long j);
static void elt_col(GEN Mk, GEN Mi, GEN q);
static void fix_cols(GEN A);
static void neg_col(GEN M);
static void reduce2(GEN A, GEN B, long k, long j, long *row);
static void swap(GEN A, GEN B, long k);
static void swap_col(GEN M, long i, long j);

GEN
myhnflll(GEN A)
{
  long m1 = 1, n1 = 1; /* alpha = m1/n1. Maybe 3/4 here ? */
  long av = avma, tetpil, do_swap,i,j,n,k, row[2];
  GEN p1,p2,B;

  n = lg(A); B = idmat(n-1); A = dummycopy(A);
  D = (GEN*) cgeti(n); lambda = (GEN**) cgetg(n,t_MAT);
  for (i=0; i<n; i++) D[i] = gun;
  for (i=1; i<n; i++)
  {
    lambda[i] = (GEN*) cgetg(n,t_COL);
    for(j=1;j<n;j++) lambda[i][j] = gzero;
  }
  k = 2;
  while (k < n)
  {
    reduce2(A,B,k,k-1,row);
    if (row[0])
      do_swap = (!row[1] || row[0] <= row[1]);
    else if (row[1])
      do_swap = 0;
    else
    { /* row[0] == row[1] == 0*/
      long av1=avma;
      p1 = addii(mulii(D[k-2],D[k]), sqri(lambda[k][k-1]));
      p1 = mulsi(n1,p1);
      p2 = mulsi(m1,sqri(D[k-1]));
      do_swap = (cmpii(p1,p2) < 0);
      avma=av1;
    }
    if (do_swap)
    {
      swap(A,B,k);
      if (k>2) k--;
    }
    else
    {
      for (i=k-2; i; i--) reduce2(A,B,k,i,row);
      k++;
    }
  }
  fix_cols(A); fix_cols(B);
  for (i=1; i<n; i++)
    if (gcmp0((GEN)A[i])) break;
  setlg(A,i); 
  tetpil=avma; p1=cgetg(3,t_VEC);
  p1[1]=lcopy(A); p1[2]=lcopy(B);
  return gerepile(av,tetpil,p1);
}

static void
fix_cols(GEN A)
{
  long i,n = lg(A);
  for (i=n>>1; i; i--) swap_col(A, i, n-i);
}

/* We assume y > 0 */
GEN
divnearest(GEN x, GEN y)
{
  GEN r, q = dvmdii(x,y,&r);
  long av = avma, s=signe(r), t;

  if (!s) { cgiv(r); return q; }
  if (s<0) r = mynegi(r);

  y = shifti(y,-1); t = cmpii(r,y);
  avma=av; cgiv(r);
  if (t < 0 || (t == 0 && s > 0)) return q;
  return addsi(s,q);
}

static void
reduce2(GEN A, GEN B, long k, long j, long *row)
{
  GEN q;
  long i, row0, row1;

  row0 = findi((GEN)A[j]);
  if (row0)
    if (signe(gcoeff(A,row0,j)) < 0)
    {
      Minus(j); neg_col((GEN)A[j]); neg_col((GEN)B[j]);
    }

  row1 = findi((GEN)A[k]);
  if (row1)
    if (signe(gcoeff(A,row1,k)) < 0)
    {
      Minus(k); neg_col((GEN)A[k]); neg_col((GEN)B[k]);
    }

  row[0]=row0; row[1]=row1;
  if (row0) 
    q = gfloor(gdiv(gcoeff(A,row0,k), gcoeff(A,row0,j)));
  else if (cmpii(gmul2n(absi(lambda[k][j]), 1), D[j]) > 0)
    q = divnearest(lambda[k][j], D[j]);
  else
    return;

  if (signe(q))
  {
    q = mynegi(q);
    elt_col((GEN)A[k],(GEN)A[j],q);
    elt_col((GEN)B[k],(GEN)B[j],q);
    lambda[k][j] = addii(lambda[k][j], mulii(q,D[j]));
    for (i=1; i<j; i++)
      if (signe(lambda[j][i]))
        lambda[k][i] = addii(lambda[k][i], mulii(q,lambda[j][i]));
  }
  if (DEBUGLEVEL)
  {
    fprintferr("A = "); output(A);
    if (DEBUGLEVEL>1) { fprintferr("B = "); output(B); }
  }
}

static void
swap(GEN A, GEN B, long k)
{
  GEN t,p1,p2;
  long i,j,n = lg(A);

  swap_col(A,k,k-1);
  swap_col(B,k,k-1);
  for (j=k-2; j; j--)
  {
    p1 = lambda[k-1][j];
    lambda[k-1][j] = lambda[k][j];
    lambda[k][j] = p1;
  }
  for (i=k+1; i<n; i++)
  {
    p1 = mulii(lambda[i][k-1], D[k]);
    p2 = mulii(lambda[i][k], lambda[k][k-1]);
    t = subii(p1,p2);

    p1 = mulii(lambda[i][k], D[k-2]);
    p2 = mulii(lambda[i][k-1], lambda[k][k-1]);
    lambda[i][k-1] = divii(addii(p1,p2), D[k-1]);

    lambda[i][k] = divii(t, D[k-1]);
  }
  p1 = mulii(D[k-2],D[k]);
  p2 = sqri(lambda[k][k-1]);
  D[k-1] = divii(addii(p1,p2), D[k-1]);
  if (DEBUGLEVEL)
  {
    if (DEBUGLEVEL>2) { fprintferr("lambda = "); output((GEN)lambda); }
    if (DEBUGLEVEL>3)
    {
      fprintferr("D = [");
      for (i=1; i<lg(A); i++)
        { bruterr(D[i],'g',-1); fprintferr(" "); }
      fprintferr("]\n"); flusherr();
    }
  }
}

static long
findi(GEN M)
{
  long i, n = lg(M);
  for (i=1; i<n; i++)
    if (signe(M[i])) return i;
  return 0;
} 

static void
Minus(long j)
{
  long k, n = lg(lambda);

  for (k=1; k<j; k++) lambda[j][k] = mynegi(lambda[j][k]);
  for (k++; k<n; k++) lambda[k][j] = mynegi(lambda[k][j]);
}

static void
neg_col(GEN M)
{
  long i;
  for (i = lg(M)-1; i; i--)
    M[i] = (long)mynegi((GEN)M[i]);
}

/* M_k = M_k + q M_i  (col operations) */
static void
elt_col(GEN Mk, GEN Mi, GEN q)
{
  long j;
  for (j = lg(Mk)-1; j; j--)
    if (signe(Mi[j]))
      Mk[j] = laddii((GEN)Mk[j], mulii(q, (GEN)Mi[j]));
}

static void
swap_col(GEN M, long i, long j)
{
  GEN u=(GEN)M[i]; M[i]=M[j]; M[j]=(long)u;
}

/* negate in place, except universal constants */
static GEN
mynegi(GEN x)
{
  static long mun[] = {evaltyp(t_INT)|evallg(3), evalsigne(-1)|evallgefint(3), 1};
  long s = signe(x);

  if (!s) return gzero;
  if (is_pm1(x))
    return (s>0)? mun: gun;
  setsigne(x,-s); return x;
}

/***********************************************************************/
/*                                                                     */
/*                 Extended GCD (Havas, Majewski, Mathews)             */
/*                                                                     */
/***********************************************************************/
static void reduce1(GEN A, GEN B, long k, long j);

GEN
extendedgcd(GEN A, long m1, long n1)
{
#if 0
  long m1 = 10001, n1 = 40000; /* alpha = m1/n1. Maybe 3/4 here ? */
#endif
  long av = avma, tetpil, do_swap,i,j,n,k;
  GEN p1,p2,B;

  n = lg(A); B = idmat(n-1); A = dummycopy(A);
  D = (GEN*) cgeti(n); lambda = (GEN**) cgetg(n,t_MAT);
  for (i=0; i<n; i++) D[i] = gun;
  for (i=1; i<n; i++)
  {
    lambda[i] = (GEN*) cgetg(n,t_COL);
    for(j=1;j<n;j++) lambda[i][j] = gzero;
  }
  k = 2;
  while (k < n)
  {
    reduce1(A,B,k,k-1);
    if (signe(A[k-1])) do_swap = 1;
    else if (!signe(A[k]))
    {
      long av1=avma;
      p1 = addii(mulii(D[k-2],D[k]), sqri(lambda[k][k-1]));
      p1 = mulsi(n1,p1);
      p2 = mulsi(m1,sqri(D[k-1]));
      do_swap = (cmpii(p1,p2) < 0);
      avma=av1;
    }
    else do_swap = 0;

    if (do_swap)
    {
      swap(A,B,k);
      if (k>2) k--;
    }
    else
    {
      for (i=k-2; i; i--) reduce1(A,B,k,i);
      k++;
    }
  }
  if (signe(A[n-1])<0)
  {
    A[n-1] = (long)mynegi((GEN)A[n-1]);
    neg_col((GEN)B[n-1]);
  }
  tetpil = avma;
  p1 = cgetg(3,t_VEC);
  p1[1]=lcopy((GEN)A[n-1]);
  p1[2]=lcopy(B);
  return gerepile(av,tetpil,p1);
}

static void
reduce1(GEN A, GEN B, long k, long j)
{
  GEN q;
  long i;

  if (signe(A[j])) 
    q = divnearest((GEN)A[k],(GEN)A[j]);
  else if (cmpii(gmul2n(absi(lambda[k][j]), 1), D[j]) > 0)
    q = divnearest(lambda[k][j], D[j]);
  else
    return;

  if (signe(q))
  {
    q = mynegi(q);
    A[k] = laddii((GEN)A[k], mulii(q,(GEN)A[j]));
    elt_col((GEN)B[k],(GEN)B[j],q);
    lambda[k][j] = addii(lambda[k][j], mulii(q,D[j]));
    for (i=1; i<j; i++)
      if (signe(lambda[j][i]))
        lambda[k][i] = addii(lambda[k][i], mulii(q,lambda[j][i]));
  }
  if (DEBUGLEVEL)
  {
    fprintferr("A = "); output(A);
    if (DEBUGLEVEL>1) { fprintferr("B = "); output(B); }
  }
}

/* A MODIFIER / VIRER. */

GEN
maxv(GEN B)
{
  GEN N,p1 = (GEN)B[1], p2 = (GEN)B[2], p3 = (GEN)B[3], vmin,v, Nmin = NULL;
  long i,j, av = avma, tetpil;

  for (i=-1;i<=1;i++)
    for (j=-1;j<=1;j++)
    {
      v = gadd(p3, gadd(gmulsg(i,p1),gmulsg(j,p2)));
      N = gnorml2(v);
      if (!Nmin || cmpii(N,Nmin) < 0)
      { 
        vmin = v; Nmin = N;
        if (DEBUGLEVEL) { fprintferr("i = %ld, j = %ld\n",i,j); }
      }
    }
  tetpil=avma; p1 = cgetg(3,t_VEC);
  p1[1]=(long)icopy(Nmin);
  p1[2]=lcopy(vmin);
  return gerepile(av,tetpil,gcopy(p1));
}

/* A VIRER. */

#define scal(x,y) gmul(gtrans((GEN)x), (GEN)y)

GEN
gs(GEN A)
{
  long n=lg(A), j,k;
  GEN p1,mu = cgetg(n,t_MAT),B = cgetg(n,t_MAT);
  B[1]=A[1];
  for (k=1; k<n; k++) mu[k]=lgetg(n,t_COL);
  for (k=1; k<n; k++)
    for (j=1; j<n; j++) mael(mu,k,j) = zero;
  for (k=2; k<n; k++)
  {
    p1 = gzero;
    for(j=1; j<k; j++)
    {
      coeff(mu,k,j) = ldiv( scal(A[k],B[j]), scal(B[j],B[j]) );
      p1 = gadd(p1, gmul(gcoeff(mu,k,j), (GEN)B[j]));
    }
    B[k]=lsub((GEN)A[k], p1);
  }
  p1 = cgetg(3,t_VEC);
  p1[1]=(long)B;
  p1[2]=(long)mu;
  return p1;
}
